/**
 ** @author: Ben Y. Zhao (ravenben@cs.berkeley.edu)
 ** @version: Release 1.1 CVS:$Revision: 1.4 $ $Date: 1999/09/10 00:12:35 $
 ** (c) 1999 by Ben Yanbin Zhao and the Regents of the University of California
 **/

package ninja.xset;

import ninja.*;

/**
 ** Class that represents the internal node structure of the Treap package
 **/
class TreapNode {
  String key;
  //  int intkey;
  //  short shortkey;
  Object value;
  int priority;
  TreapNode left, right;
  //  short treapType = Treap.STRINGTREAP;

  /**
   ** Constructor method
   ** @param k String representing the key
   ** @param v Object associated with the key
   **/  
  TreapNode(String k, Object v) {
    key = k;
    value = v;
    left = right = null;
    priority = (int) (65536.0*Math.random());
  }
  /*  TreapNode(int k, Object v) {
      intkey = k;
      value = v;
      left = right = null;
      priority = (int) (65536.0*Math.random());
      treapType = Treap.INTTREAP;
      }
      TreapNode(short k, Object v) {
      shortkey = k;
      value = v;
      left = right = null;
      priority = (int) (65536.0*Math.random());
    treapType = Treap.SHORTTREAP;
    }*/
  /**
   ** Constructor method with manually set priority
   ** @param p priority value
   ** @param k String representing the key
   ** @param v Object associated with the key
   **/
  TreapNode(String k, Object v, int p) {
    key = k; priority = p; value = v; left = right = null;
  }
  /*  TreapNode(int k, Object v, int p) {
      intkey = k; priority = p; value = v; left = right = null;
    treapType = Treap.INTTREAP;
    }
    TreapNode(short k, Object v, int p) {
    shortkey = k; priority = p; value = v; left = right = null;
    treapType = Treap.SHORTTREAP;
    }*/
  
  /**
   ** Method to compute the average depth of the treap
   ** @return DepthCounter object
   **/
  DepthCounter averageDepth() {
    if (left == null) {
      if (right == null) {
	return new DepthCounter(1, 1);
      } else {
	DepthCounter d = right.averageDepth();
	d.totalDepth += d.nodes + 1;
	d.nodes++;
	return d;
      }
    }
    
    DepthCounter d = left.averageDepth();
    d.totalDepth += d.nodes + 1;
    d.nodes++;
    
    if (right == null) return d;
    
    DepthCounter r = right.averageDepth();
    d.totalDepth += (r.totalDepth + r.nodes);
    d.nodes += r.nodes;
    
    return d;
  }
  
  /**
   ** Method to verify treap invariants
   **/
  void check() {
    // verify key order
    if (left != null) Ninja.assert(left.key.compareTo(key) < 0);
    if (right != null) Ninja.assert(key.compareTo(right.key) < 0);
    
    // verify heap order
    if (left != null) Ninja.assert(left.priority >= priority);
    if (right != null) Ninja.assert(right.priority >= priority);
  }
  
  /**
   **   delete node with key k
   **   result is a (pass-by-ref) object to hold the value
   **     of the deleted node
   **   returns the (potentially) new root of this subtree
   **     the caller is expected to update their pointer(s)
   **     to this subtree
   ** @param k String representing string key
   ** @param result TreapNode that will hold the value of the deleted node
   ** @return TreapNode root of the subtree, could be new, need to be updated
   **/
  TreapNode delete(String k, TreapNode result) {
    int c = k.compareTo(key);
    
    if (c < 0) {
      if (left == null) {
	result.value = null;
      } else {	
	left = left.delete(k, result);
      }
      return this;
    }	
    
    if (c > 0) {
      if (right == null) {
	result.value = null;
      } else {	
	right = right.delete(k, result);
      }	
      return this;
    }
    
    Ninja.assert(k.equals(key));
    
    // delete this node, save value first
    
    if (left == null) {
      // replace with right subtree
      result.value = value;
      return right;
    }
    if (right == null) {
      // replace with left subtree
      result.value = value;
      return left;
    }
    
    // neither subtree is empty, so rotate target node down
    TreapNode newRoot;
    
    if (left.priority > right.priority) {
      newRoot = rotateLeft();
      left = left.delete(k, result);
    } else {
      newRoot = rotateRight();
      right = right.delete(k, result);
    }
    return newRoot;
  }	
  
  /**
   ** Method to compute left - right
   ** @param t
   ** @param l left Treap
   ** @param r right Treap
   ** @return TreapNode that is the root of the new difference treap
   **/
  static TreapNode difference(Treap t, TreapNode l, TreapNode r) {
    TreapNode result, root, split, left, right;
    
    if (r == null) return l;
    if (l == null) return null;
    
    if (l.priority < r.priority) {
      root = r;
      split = l.split(r.key);
      left = difference(t, split.left, r.left);
      right = difference(t, split.right, r.right);
    } else {
      root = l;
      split = r.split(l.key);
      left = difference(t, l.left, split.left);
      right = difference(t, l.right, split.right);
    }
    
    if (split.key != null) {
      // duplicate node
      t.nodeCount--;
      result = join(left, right);
    } else if (root == r) {
      // no duplicate, but root not part of the set
      result = join(left, right);
    } else {
      // no duplicate, and root is in the set
      result = root;
      result.left = left;
      result.right = right;
    }		
    return result;
  }

  /**
   ** Method to insert (k,v) into treap
   ** assumes no duplicates, and raises an exception if there is...
   ** @param k String representing the key
   ** @param v Object associated with the key
   ** @return TreapNode that is the possibly new root of the current subTreap
   **/
  TreapNode insert(String k, Object v) throws DuplicateException {
    int c = k.compareTo(key);
    if (c < 0) {
      if (left == null) {
	left = new TreapNode(k,v);
      } else {
	left = left.insert(k, v);
      }
      if (priority > left.priority) {
	return rotateRight();
      }
    } else if (c > 0) {
      if (right == null) {
	right = new TreapNode(k,v);
      } else {
	right = right.insert(k, v);
      }
      if (priority > right.priority)  {
	return rotateLeft();
      }
    } else {
      throw new DuplicateException(k);
    }
    return this;
  }
  
  /**
   **   Compute the non-destructive intersection of two treaps
   **   For each node in common, compute the merged (output) node
   **   by invoking intersectionMerge on the two nodes.
   **/
  static TreapNode intersection(Treap t, TreapNode l, TreapNode r) {
    TreapNode result, left, right, split, root;
    
    if (l == null || r == null) return null;
    
    if (l.priority < r.priority) {
      root = r;
      split = l.split(r.key);
    } else {
      root = l;
      split = r.split(l.key);
    }		
    left = intersection(t, split.left, root.left);
    right = intersection(t, split.right, root.right);
    
    if (split.key != null) {
      // duplicate node
      t.nodeCount++;
      result = new TreapNode(root.key, root.value, root.priority);
      if (t.merge != null)
	result = t.merge.intersectionMerge(result, split);
      result.left = left;
      result.right = right;
    } else {
      result = join(left, right);
    }		
    
    return result;
  }

  /**
   **   join(l, r)  returns a new combined tree out of tree l and r
   **     It assumes that max(l) < min(r)
   **/
  static TreapNode join(TreapNode l, TreapNode r) {
    TreapNode root;
    
    if (l == null) return r;
    if (r == null) return l;
    
    if (l.priority < r.priority) {
      root = new TreapNode(l.key, l.value, l.priority);
      root.left = l.left;
      root.right = join(l.right, r);
    } else {
      root = new TreapNode(r.key, r.value, r.priority);
      root.left = join(l, r.left);
      root.right = r.right;
    }	
    return root;
  }

  /**
   ** get the value for key k
   ** return null if it is not in the treap
   ** @param k String representation of the key
   ** @return Object associated with key
   **/
  // OLD LOOKUP using RECURSION
  /*  Object lookup(String k) {
    int c = k.compareTo(key);
    if (c < 0) {
      if (left == null) return null;
      return left.lookup(k);
    } else if (c > 0) {
      if (right == null) return null;
      return right.lookup(k);
    } else {
      return value;
    }
  }*/
  Object lookup(String k) {
    int c;
    TreapNode node = this;
    while (true) {
      c = k.compareTo(node.key);
      if (c < 0) {
	if (node.left == null) return null;
	node = node.left;
      } else if (c > 0) {
	if (node.right == null) return null;
	node = node.right;
      } else return node.value;
    }
  }

  TreapNode findNode(String k, boolean [] b) {
    int c;
    TreapNode node = this;
    while (true) {
      System.out.println("Entering lookup node: key "+node.key);
      c = k.compareTo(node.key);
      if (c < 0) {
	if (node.left == null) {   // hit edge of treap, return edgenode
	  b[0] = false;            // but sideeffect inexactness
	  return node;
	}
	node = node.left;
      } else if (c > 0) {
	if (node.right == null) {  // hit edge of treap, return edgenode
	  b[0] = false;            // but sideeffect inexactness
	  return node;
	}
	node = node.right;
      } else return node;
    }
  }
  /*
  TreapNode findNode(int k, boolean [] b) {
    int intkey = Integer.parseInt(key);
    TreapNode node = this;
    while (true) {
      if (k < intkey) {
	if (node.left == null) {  // hit edge of treap, return edgenode
	  b[0] = false;           // but sideeffect inexactness
	  return this;
	}
	node = node.left;
      } else if (k > intkey) {    
	if (node.right == null) { // hit edge of treap, return edgenode
	  b[0] = false;           // but sideeffect inexactness
	  return this;
	}
	node = node.right;
      } else return this;
    }
  }
  TreapNode findNode(float k, boolean [] b) {
    float fkey = (new Float(key)).floatValue();
    TreapNode node = this;
    while (true) {
      if (k < fkey) {
	if (node.left == null) {  // hit edge of treap, return edgenode
	  b[0] = false;           // but sideeffect inexactness
	  return this;
	}
	node = node.left;
      } else if (k > fkey) {
	if (node.right == null) { // hit edge of treap, return edgenode
	  b[0] = false;           // but sideeffect inexactness
	  return this;
	}
	node = node.right;
      } else return this;
    }
  }
  **/
  /**
   ** Compute the maximum depth of the treap
   ** @return integer maximum depth of this subTreap
   **/
  int maxDepth() {
    if (left == null) {
      if (right == null) {
	return 1;
      }
      return right.maxDepth() + 1;
    }
    if (right == null) {
      return left.maxDepth() + 1;
    }
    
    int l = left.maxDepth();
    int r = right.maxDepth();
    if (l > r) return l+1;
    return r+1;
  }
  
  /**
   **  print treap with indentation
   ** @param depth Integer number - amount of indentation
   **/
  // package access only (not public or private)
  void printIndent(int depth) {
    if (right != null) 	right.printIndent(depth + 1);
    for (int i=0; i<depth; i++) {
      System.out.print("   ");
    }
    System.out.println(key + "=" + value + " [" + priority + "]");
    if (left != null) left.printIndent(depth+1);	
    return;
  }

  /**
   **   perform a left rotation to fix priority heap
   **/
  private TreapNode rotateLeft() {
    TreapNode x = right;
    right = x.left;
    x.left = this;
    return x;
  }
  
  /**
   **  Rotate subtree right. Returns new root of the subtree.
   **/
  private TreapNode rotateRight() {
    TreapNode x = left;
    left = x.right;
    x.right = this;
    return x;
  }

  /**
   ** Method to split a treap on a key
   **/
  TreapNode split(String k) {
    TreapNode split;
    int c = k.compareTo(key);
    
    if (c < 0) {
      if (left == null) {
	// split key not in tree
	split = new TreapNode(null, null);
	split.left = null;
	split.right = this;
      } else {
	// split key is less than root, so split left subtree
	split = left.split(k);
	TreapNode top = new TreapNode(key, value, priority);
	top.left = split.right;
	top.right = right;
	split.right = top;
      }
    } else if (c > 0) {
      if (right == null) {
	// split key not in treap
	split = new TreapNode(null, null);
	split.left = this;
	split.right = null;
      } else {
	// split key is greater than root, so split right subtree
	split = right.split(k);
	TreapNode top = new TreapNode(key, value, priority);
	top.left = left;
	top.right = split.left;
	split.left = top;
      }	
    } else {
      // this key is split node
      split = new TreapNode(key, value, priority);
      split.left = left;
      split.right = right;
    }
    return split;		
  }

  /**
   **   Compute l + r  (as sets)
   **   Uses t.merge to figure out how to merge duplicates
   **/
  static TreapNode union(Treap t, TreapNode l, TreapNode r) {
    TreapNode result, split, root;
	
    if (l == null) return r;
    if (r == null) return l;
    
    if (l.priority < r.priority) {
      root = r;
      split = l.split(r.key);
    } else {
      root = l;
      split = r.split(l.key);
    }		
    result = new TreapNode(root.key, root.value, root.priority);
    
    if (split.key != null) {
      // duplicate node; correct overall nodeCount
      t.nodeCount--;
      if (t.merge != null)
	result = t.merge.unionMerge(result, split);
    }
    result.left = union(t, split.left, root.left);
    result.right = union(t, split.right, root.right);	
    return result;
  }
}
